#!/usr/bin/env python
# -*- coding: utf-8; -*-

# 2018-03-07 David Goncalves <david@lestat.st>
#            Python class to communicate with an LCM1602 chip on I2C bus.
#            This is an LCD i2c driver based on the PCF8574 chip (8 bit I/O expander)

import smbus
import sys
import time


class lcm1602:

    # Communication bits
    BIT_RS = 1              # Register Select (0=command 1=data)
    BIT_RW = 1 << 1         # Read/Write bit (0=write 1=read)
    BIT_EN = 1 << 2         # Enable bit
    BIT_BL = 1 << 3         # Backlight

    # Commands
    CMD_CLEAR_DISPLAY = 0x01
    CMD_RETURN_HOME = 0x02
    CMD_ENTRY_MODE = 0x04
    CMD_DISPLAY_CONTROL = 0x08
    CMD_SET_DDRAM_ADDRESS = 0x80

    # Flags for display control
    DISPLAY_BLINK = 0x01
    DISPLAY_CURSOR = 0x02
    DISPLAY_ON = 0x04

    # Flags for entry mode
    ENTRY_MOVE_RIGHT = 0x02
    ENTRY_MOVE_LEFT = 0x00
    ENTRY_SHIFT_DISPLAY = 0x01

    rows = 2
    backlight = True

    lcd_line_offset = [0x00, 0x40, 0x14, 0x54]  # Addresses of first bytes of LCD lines

    def __init__(self, address=0x27, bus=1, rows=2, cols=16, backlight=True, debug=False):
        self._debug = debug
        self.i2c_addr = address
        self.i2c_bus = smbus.SMBus(bus)

        self._debug_message("Accessing LCM1602 at address 0x%x on bus %s with %s rows" % (address, bus, rows))

        if rows not in range(1, 5):
            raise ValueError("Invalid row count, range must be from 1 to 4")

        self._debug_message("Switching to 4bit mode")
        for delay in [0.005, 0.005, 0.0001]:
            self._write_data(0, [(0x3 << 4) | (self.BIT_BL if self.backlight is True else 0)])
            time.sleep(delay)

        self._write_data(0, [(0x2 << 4) | (self.BIT_BL if self.backlight is True else 0)])

        self.rows = rows
        self.cols = cols

    def _debug_message(self, message="", error=False):
        """
        Prints debug messages to console when debuging mode is activated
        """

        if (self._debug is True) or (error is True):
            print("%s %s" % (("[DEBUG]" if error is False else "[ERROR]"), message))
            sys.stdout.flush()
            if (error is True):
                raise Exception(message)

    def _write_data(self, address, data):
        """
        Private method to write data to the LCM1602
        """

        self._debug_message("> Writing %s bytes to register 0x%x [%s]" % (len(data), address, ' '.join(["0x%02X" % x for x in data])))
        self.i2c_bus.write_i2c_block_data(self.i2c_addr, address, data)

    def _read_data(self, address, size):
        """
        Private method to read data from the LCM1602
        """

        self._debug_message("> Reading %s bytes from register 0x%x" % (size, address))
        data = self.i2c_bus.read_i2c_block_data(self.i2c_addr, address, size)
        self._debug_message("> LCM1602 response : [%s]" % ' '.join(["0x%02X" % x for x in data]))

        return(data)

    def _write_lcd(self, data, rs=False):
        """
        Private method to write to the LCD controler by splitting data into 4 bit nibbles
        and dealing with R/W and CS bits to communicate.

        If 'rs' is True, the BIT_RS will be set to 1 (0=command 1=data)
        """

        # Separate byte into 2 nibbles and include the backlight status
        high_nibble = (data & 0xf0) | (self.BIT_BL if self.backlight is True else 0) | (self.BIT_RS if rs is True else 0)
        low_nibble = (data & 0x0f) << 4 | (self.BIT_BL if self.backlight is True else 0) | (self.BIT_RS if rs is True else 0)

        self._write_data(0, [(high_nibble | self.BIT_EN) & 0xff])
        time.sleep(0.000001)                               # Enable pulse > 450ns
        self._write_data(0, [(high_nibble & ~ self.BIT_EN) & 0xff])
        time.sleep(0.00005)                                # Settle > 37µs

        self._write_data(0, [(low_nibble | self.BIT_EN) & 0xff])
        time.sleep(0.000001)                               # Enable pulse > 450ns
        self._write_data(0, [(low_nibble & ~ self.BIT_EN) & 0xff])
        time.sleep(0.00051)                                # Settle > 37µs

    def SetBacklight(self, status=True):
        """
        Set backlight on (True) or off (False)
        """

        self._debug_message("Setting backlight to %s" % ("on" if status is True else "off"))
        self._write_data(0, [(self.BIT_BL if status is True else 0)])
        self.backlight = status

    def SetDisplayFlags(self, flags=DISPLAY_ON):
        """
        Set display flags (use DISPLAY_* constants to configure)
        """

        self._debug_message("Setting display flags")
        self._write_lcd(self.CMD_DISPLAY_CONTROL | flags)

    def SetEntryMode(self, mode=ENTRY_MOVE_RIGHT):
        """
        Set entry mode (use ENTRY_* constants to configure)
        """

        self._debug_message("Setting entry mode")
        self._write_lcd(self.CMD_ENTRY_MODE | mode)

    def ClearDisplay(self):
        """
        Clears the display
        """

        self._debug_message("Clearing display")
        self._write_lcd(self.CMD_RETURN_HOME)
        self._write_lcd(self.CMD_CLEAR_DISPLAY)

    def DisplayText(self, text="", row=1, col=1):
        """
        Display a text at row/col position
        """

        self._debug_message("Displaying '%s' at row=%s col=%s" % (text, row, col))

        if row < 1 or row > self.rows:
            raise ValueError("Row out of range '%s'" % row)

        if col < 1 or col > self.cols:
            raise ValueError("Column out of range '%s'" % col)

        char_address = self.lcd_line_offset[row - 1] + (col - 1)
        self._write_lcd(self.CMD_SET_DDRAM_ADDRESS | char_address)

        for c in text:
            self._write_lcd(ord(c), rs=True)
